@@ -17,12 +17,6 @@
  *    INTERNAL TYPES DECLARATIONS   *
  *----------------------------------*/
 
-/*
- * Type for keeping intermediate values of TrieIndex.
- * Must be bigger than TrieIndex, so that overflow can be easily detected.
- */
-typedef int32           TrieIndexInt;
-
 typedef struct _Symbols Symbols;
 
 struct _Symbols {
@@ -45,15 +39,15 @@ static void         symbols_add (Symbols *syms, TrieChar c);
 #define da_get_free_list(d)      (1)
 
 static Bool         da_check_free_cell (DArray         *d,
-                                        TrieIndexInt    s);
+                                        TrieIndex       s);
 
 static Bool         da_has_children    (DArray         *d,
                                         TrieIndex       s);
 
-static Symbols *    da_output_symbols  (DArray         *d,
+static Symbols *    da_output_symbols  (const DArray   *d,
                                         TrieIndex       s);
 
-static TrieChar *   da_get_state_key   (DArray         *d,
+static TrieChar *   da_get_state_key   (const DArray   *d,
                                         TrieIndex       state);
 
 static TrieIndex    da_find_free_base  (DArray         *d,
@@ -68,7 +62,7 @@ static void         da_relocate_base   (DArray         *d,
                                         TrieIndex       new_base);
 
 static Bool         da_extend_pool     (DArray         *d,
-                                        TrieIndexInt    to_index);
+                                        TrieIndex       to_index);
 
 static void         da_alloc_cell      (DArray         *d,
                                         TrieIndex       cell);
@@ -76,10 +70,10 @@ static void         da_alloc_cell      (DArray         *d,
 static void         da_free_cell       (DArray         *d,
                                         TrieIndex       cell);
 
-static Bool         da_enumerate_recursive (DArray     *d,
-                                            TrieIndex   state,
-                                            DAEnumFunc  enum_func,
-                                            void       *user_data);
+static Bool         da_enumerate_recursive (const DArray   *d,
+                                            TrieIndex       state,
+                                            DAEnumFunc      enum_func,
+                                            void           *user_data);
 
 /* ==================== BEGIN IMPLEMENTATION PART ====================  */
 
@@ -146,19 +140,16 @@ typedef struct {
 struct _DArray {
     TrieIndex   num_cells;
     DACell     *cells;
-
-    FILE       *file;
-    Bool        is_dirty;
 };
 
 /*-----------------------------*
  *    METHODS IMPLEMENTAIONS   *
  *-----------------------------*/
 
-#define DA_SIGNATURE 0xDAFD
+#define DA_SIGNATURE 0xDAFCDAFC
 
 /* DA Header:
- * - Cell 0: SIGNATURE, 1
+ * - Cell 0: SIGNATURE, number of cells
  * - Cell 1: free circular-list pointers
  * - Cell 2: root node
  * - Cell 3: DA pool begin
@@ -166,89 +157,88 @@ struct _DArray {
 #define DA_POOL_BEGIN 3
 
 DArray *
-da_open (const char *path, const char *name, TrieIOMode mode)
+da_new ()
 {
     DArray     *d;
-    TrieIndex   i;
 
     d = (DArray *) malloc (sizeof (DArray));
+    if (!d)
+        return NULL;
 
-    d->file = file_open (path, name, ".br", mode);
-    if (!d->file)
-        goto exit1;
-
-    /* init cells data */
-    d->num_cells = file_length (d->file) / 4;
-    if (0 == d->num_cells) {
-        d->num_cells = DA_POOL_BEGIN;
-        d->cells     = (DACell *) malloc (d->num_cells * sizeof (DACell));
-        if (!d->cells)
-            goto exit2;
-        d->cells[0].base = DA_SIGNATURE;
-        d->cells[0].check = 1;
-        d->cells[1].base = -1;
-        d->cells[1].check = -1;
-        d->cells[2].base = DA_POOL_BEGIN;
-        d->cells[2].check = 0;
-        d->is_dirty = TRUE;
-    } else {
-        d->cells     = (DACell *) malloc (d->num_cells * sizeof (DACell));
-        if (!d->cells)
-            goto exit2;
-        file_read_int16 (d->file, &d->cells[0].base);
-        file_read_int16 (d->file, &d->cells[0].check);
-        if (DA_SIGNATURE != (uint16) d->cells[0].base)
-            goto exit3;
-        for (i = 1; i < d->num_cells; i++) {
-            file_read_int16 (d->file, &d->cells[i].base);
-            file_read_int16 (d->file, &d->cells[i].check);
-        }
-        d->is_dirty  = FALSE;
-    }
+    d->num_cells = DA_POOL_BEGIN;
+    d->cells     = (DACell *) malloc (d->num_cells * sizeof (DACell));
+    if (!d->cells)
+        goto exit_da_created;
+    d->cells[0].base = DA_SIGNATURE;
+    d->cells[0].check = d->num_cells;
+    d->cells[1].base = -1;
+    d->cells[1].check = -1;
+    d->cells[2].base = DA_POOL_BEGIN;
+    d->cells[2].check = 0;
 
     return d;
 
-exit3:
-    free (d->cells);
-exit2:
-    fclose (d->file);
-exit1:
+exit_da_created:
     free (d);
     return NULL;
 }
 
-int
-da_close (DArray *d)
+DArray *
+da_read (FILE *file)
 {
-    int ret;
+    long        save_pos;
+    DArray     *d = NULL;
+    TrieIndex   n;
 
-    if (0 != (ret = da_save (d)))
-        return ret;
-    if (0 != (ret = fclose (d->file)))
-        return ret;
+    /* check signature */
+    save_pos = ftell (file);
+    if (!file_read_int32 (file, &n) || DA_SIGNATURE != (uint32) n) {
+        fseek (file, save_pos, SEEK_SET);
+        return NULL;
+    }
+
+    d = (DArray *) malloc (sizeof (DArray));
+    if (!d)
+        return NULL;
+
+    /* read number of cells */
+    file_read_int32 (file, &d->num_cells);
+    d->cells     = (DACell *) malloc (d->num_cells * sizeof (DACell));
+    if (!d->cells)
+        goto exit_da_created;
+    d->cells[0].base = DA_SIGNATURE;
+    d->cells[0].check= d->num_cells;
+    for (n = 1; n < d->num_cells; n++) {
+        file_read_int32 (file, &d->cells[n].base);
+        file_read_int32 (file, &d->cells[n].check);
+    }
+
+    return d;
+
+exit_da_created:
+    free (d);
+    return NULL;
+}
+
+void
+da_free (DArray *d)
+{
     free (d->cells);
     free (d);
-
-    return 0;
 }
 
 int
-da_save (DArray *d)
+da_write (const DArray *d, FILE *file)
 {
     TrieIndex   i;
 
-    if (!d->is_dirty)
-        return 0;
-
-    rewind (d->file);
     for (i = 0; i < d->num_cells; i++) {
-        if (!file_write_int16 (d->file, d->cells[i].base) ||
-            !file_write_int16 (d->file, d->cells[i].check))
+        if (!file_write_int32 (file, d->cells[i].base) ||
+            !file_write_int32 (file, d->cells[i].check))
         {
             return -1;
         }
     }
-    d->is_dirty = FALSE;
 
     return 0;
 }
@@ -280,7 +270,6 @@ da_set_base (DArray *d, TrieIndex s, TrieIndex val)
 {
     if (s < d->num_cells) {
         d->cells[s].base = val;
-        d->is_dirty = TRUE;
     }
 }
 
@@ -289,12 +278,11 @@ da_set_check (DArray *d, TrieIndex s, TrieIndex val)
 {
     if (s < d->num_cells) {
         d->cells[s].check = val;
-        d->is_dirty = TRUE;
     }
 }
 
 Bool
-da_walk (DArray *d, TrieIndex *s, TrieChar c)
+da_walk (const DArray *d, TrieIndex *s, TrieChar c)
 {
     TrieIndex   next;
 
@@ -309,18 +297,21 @@ da_walk (DArray *d, TrieIndex *s, TrieChar c)
 TrieIndex
 da_insert_branch (DArray *d, TrieIndex s, TrieChar c)
 {
-    TrieIndexInt    base, next;
+    TrieIndex   base, next;
 
     base = da_get_base (d, s);
 
     if (base > 0) {
-        next = da_get_base (d, s) + c;
+        next = base + c;
 
         /* if already there, do not actually insert */
         if (da_get_check (d, next) == s)
             return next;
 
-        if (!da_check_free_cell (d, next)) {
+        /* if (base + c) > TRIE_INDEX_MAX which means 'next' is overflow,
+         * or cell [next] is not free, relocate to a free slot
+         */
+        if (base > TRIE_INDEX_MAX - c || !da_check_free_cell (d, next)) {
             Symbols    *symbols;
             TrieIndex   new_base;
 
@@ -359,7 +350,7 @@ da_insert_branch (DArray *d, TrieIndex s, TrieChar c)
 
 static Bool
 da_check_free_cell (DArray         *d,
-                    TrieIndexInt    s)
+                    TrieIndex       s)
 {
     return da_extend_pool (d, s) && da_get_check (d, s) < 0;
 }
@@ -369,7 +360,7 @@ da_has_children    (DArray         *d,
                     TrieIndex       s)
 {
     TrieIndex   base;
-    uint16      c, max_c;
+    TrieIndex   c, max_c;
 
     base = da_get_base (d, s);
     if (TRIE_INDEX_ERROR == base || base < 0)
@@ -385,12 +376,12 @@ da_has_children    (DArray         *d,
 }
 
 static Symbols *
-da_output_symbols  (DArray         *d,
+da_output_symbols  (const DArray   *d,
                     TrieIndex       s)
 {
     Symbols    *syms;
     TrieIndex   base;
-    uint16      c, max_c;
+    TrieIndex   c, max_c;
 
     syms = symbols_new ();
 
@@ -405,7 +396,7 @@ da_output_symbols  (DArray         *d,
 }
 
 static TrieChar *
-da_get_state_key   (DArray         *d,
+da_get_state_key   (const DArray   *d,
                     TrieIndex       state)
 {
     TrieChar   *key;
@@ -447,7 +438,7 @@ da_find_free_base  (DArray         *d,
                     const Symbols  *symbols)
 {
     TrieChar        first_sym;
-    TrieIndexInt    s;
+    TrieIndex       s;
 
     /* find first free cell that is beyond the first symbol */
     first_sym = symbols_get (symbols, 0);
@@ -488,7 +479,12 @@ da_fit_symbols     (DArray         *d,
     int         i;
 
     for (i = 0; i < symbols_num (symbols); i++) {
-        if (!da_check_free_cell (d, base + symbols_get (symbols, i)))
+        TrieChar    sym = symbols_get (symbols, i);
+
+        /* if (base + sym) > TRIE_INDEX_MAX which means it's overflow,
+         * or cell [base + sym] is not free, the symbol is not fit.
+         */
+        if (base > TRIE_INDEX_MAX - sym || !da_check_free_cell (d, base + sym))
             return FALSE;
     }
     return TRUE;
@@ -524,7 +520,7 @@ da_relocate_base   (DArray         *d,
          */
         /* preventing the case of TAIL pointer */
         if (old_next_base > 0) {
-            uint16      c, max_c;
+            TrieIndex   c, max_c;
 
             max_c = MIN_VAL (TRIE_CHAR_MAX, TRIE_INDEX_MAX - old_next_base);
             for  (c = 0; c < max_c; c++) {
@@ -545,7 +541,7 @@ da_relocate_base   (DArray         *d,
 
 static Bool
 da_extend_pool     (DArray         *d,
-                    TrieIndexInt    to_index)
+                    TrieIndex       to_index)
 {
     TrieIndex   new_begin;
     TrieIndex   i;
@@ -574,6 +570,9 @@ da_extend_pool     (DArray         *d,
     da_set_check (d, to_index, -da_get_free_list (d));
     da_set_base (d, da_get_free_list (d), -to_index);
 
+    /* update header cell */
+    d->cells[0].check = d->num_cells;
+
     return TRUE;
 }
 
@@ -630,16 +629,16 @@ da_free_cell       (DArray         *d,
 }
 
 Bool
-da_enumerate (DArray *d, DAEnumFunc enum_func, void *user_data)
+da_enumerate (const DArray *d, DAEnumFunc enum_func, void *user_data)
 {
     return da_enumerate_recursive (d, da_get_root (d), enum_func, user_data);
 }
 
 static Bool
-da_enumerate_recursive (DArray     *d,
-                        TrieIndex   state,
-                        DAEnumFunc  enum_func,
-                        void       *user_data)
+da_enumerate_recursive (const DArray   *d,
+                        TrieIndex       state,
+                        DAEnumFunc      enum_func,
+                        void           *user_data)
 {
     Bool        ret;
     TrieIndex   base;
