@@ -9,13 +9,7 @@
 #define __FILEUTILS_H
 
 #include <stdio.h>
-
-#include "triedefs.h"
-
-FILE * file_open (const char *dir, const char *name, const char *ext,
-                  TrieIOMode mode);
-
-long   file_length (FILE *file);
+#include <datrie/typedefs.h>
 
 Bool   file_read_int32 (FILE *file, int32 *o_val);
 Bool   file_write_int32 (FILE *file, int32 val);
