@@ -9,6 +9,9 @@
 #include <string.h>
 
 #include "trie.h"
+#include "fileutils.h"
+#include "alpha-map.h"
+#include "alpha-map-private.h"
 #include "darray.h"
 #include "tail.h"
 
@@ -16,18 +19,21 @@
  * @brief Trie structure
  */
 struct _Trie {
-    DArray  *da;
-    Tail    *tail;
+    AlphaMap   *alpha_map;
+    DArray     *da;
+    Tail       *tail;
+
+    Bool        is_dirty;
 };
 
 /**
  * @brief TrieState structure
  */
 struct _TrieState {
-    Trie      *trie;        /**< the corresponding trie */
-    TrieIndex  index;       /**< index in double-array/tail structures */
-    short      suffix_idx;  /**< suffix character offset, if in suffix */
-    short      is_suffix;   /**< whether it is currently in suffix part */
+    const Trie *trie;       /**< the corresponding trie */
+    TrieIndex   index;      /**< index in double-array/tail structures */
+    short       suffix_idx; /**< suffix character offset, if in suffix */
+    short       is_suffix;  /**< whether it is currently in suffix part */
 };
 
 /*------------------------*
@@ -38,10 +44,10 @@ struct _TrieState {
 #define trie_da_get_tail_index(da,s)   (-da_get_base ((da), (s)))
 #define trie_da_set_tail_index(da,s,v) (da_set_base ((da), (s), -(v)))
 
-static TrieState * trie_state_new (Trie      *trie,
-                                   TrieIndex  index,
-                                   short      suffix_idx,
-                                   short      is_suffix);
+static TrieState * trie_state_new (const Trie *trie,
+                                   TrieIndex   index,
+                                   short       suffix_idx,
+                                   short       is_suffix);
 
 static Bool        trie_branch_in_branch (Trie           *trie,
                                           TrieIndex       sep_node,
@@ -58,7 +64,7 @@ static Bool        trie_branch_in_tail   (Trie           *trie,
  *-----------------------*/
 
 Trie *
-trie_open (const char *path, const char *name, TrieIOMode mode)
+trie_new (const AlphaMap *alpha_map)
 {
     Trie *trie;
 
@@ -66,50 +72,112 @@ trie_open (const char *path, const char *name, TrieIOMode mode)
     if (!trie)
         return NULL;
 
-    if (NULL == (trie->da   = da_open (path, name, mode)))
-        goto exit1;
-    if (NULL == (trie->tail = tail_open (path, name, mode)))
-        goto exit2;
+    trie->alpha_map = alpha_map_clone (alpha_map);
+    if (!trie->alpha_map)
+        goto exit_trie_created;
 
+    trie->da = da_new ();
+    if (!trie->da)
+        goto exit_alpha_map_created;
+
+    trie->tail = tail_new ();
+    if (!trie->tail)
+        goto exit_da_created;
+ 
+    trie->is_dirty = TRUE;
     return trie;
 
-exit2:
-    da_close (trie->da);
-exit1:
+exit_da_created:
+    da_free (trie->da);
+exit_alpha_map_created:
+    alpha_map_free (trie->alpha_map);
+exit_trie_created:
     free (trie);
     return NULL;
 }
 
-int
-trie_close (Trie *trie)
+Trie *
+trie_new_from_file (const char *path)
 {
+    Trie       *trie;
+    FILE       *trie_file;
+
+    trie_file = fopen (path, "r");
+    if (!trie_file)
+        return NULL;
+
+    trie = (Trie *) malloc (sizeof (Trie));
     if (!trie)
-        return -1;
+        goto exit_file_openned;
 
-    if (da_close (trie->da) != 0)
-        return -1;
+    if (NULL == (trie->alpha_map = alpha_map_read_bin (trie_file)))
+        goto exit_trie_created;
+    if (NULL == (trie->da   = da_read (trie_file)))
+        goto exit_alpha_map_created;
+    if (NULL == (trie->tail = tail_read (trie_file)))
+        goto exit_da_created;
 
-    if (tail_close (trie->tail) != 0)
-        return -1;
+    fclose (trie_file);
+    trie->is_dirty = FALSE;
+    return trie;
 
+exit_da_created:
+    da_free (trie->da);
+exit_alpha_map_created:
+    alpha_map_free (trie->alpha_map);
+exit_trie_created:
     free (trie);
+exit_file_openned:
+    fclose (trie_file);
+    return NULL;
+}
 
-    return 0;
+void
+trie_free (Trie *trie)
+{
+    alpha_map_free (trie->alpha_map);
+    da_free (trie->da);
+    tail_free (trie->tail);
+    free (trie);
 }
 
 int
-trie_save (Trie *trie)
+trie_save (Trie *trie, const char *path)
 {
-    if (!trie)
+    FILE *file;
+    int   res = 0;
+
+    file = fopen (path, "w+");
+    if (!file)
         return -1;
 
-    if (da_save (trie->da) != 0)
-        return -1;
+    if (alpha_map_write_bin (trie->alpha_map, file) != 0) {
+        res = -1;
+        goto exit_file_openned;
+    }
 
-    if (tail_save (trie->tail) != 0)
-        return -1;
+    if (da_write (trie->da, file) != 0) {
+        res = -1;
+        goto exit_file_openned;
+    }
 
+    if (tail_write (trie->tail, file) != 0) {
+        res = -1;
+        goto exit_file_openned;
+    }
+
+    trie->is_dirty = FALSE;
     return 0;
+
+exit_file_openned:
+    fclose (file);
+    return res;
+}
+
+Bool
+trie_is_dirty (const Trie *trie)
+{
+    return trie->is_dirty;
 }
 
 
@@ -118,28 +186,36 @@ trie_save (Trie *trie)
  *------------------------------*/
 
 Bool
-trie_retrieve (Trie *trie, const TrieChar *key, TrieData *o_data)
+trie_retrieve (const Trie *trie, const AlphaChar *key, TrieData *o_data)
 {
     TrieIndex        s;
     short            suffix_idx;
-    const TrieChar  *p;
-    size_t           len;
+    const AlphaChar *p;
 
     /* walk through branches */
     s = da_get_root (trie->da);
     for (p = key; !trie_da_is_separate (trie->da, s); p++) {
-        if (!da_walk (trie->da, &s, *p))
+        if (!da_walk (trie->da, &s,
+                      alpha_map_char_to_trie (trie->alpha_map, *p)))
+        {
             return FALSE;
-        if ('\0' == *p)
+        }
+        if (0 == *p)
             break;
     }
 
     /* walk through tail */
     s = trie_da_get_tail_index (trie->da, s);
     suffix_idx = 0;
-    len = strlen ((const char *) p) + 1;    /* including null-terminator */
-    if (tail_walk_str (trie->tail, s, &suffix_idx, p, len) != len)
-        return FALSE;
+    for ( ; ; p++) {
+        if (!tail_walk_char (trie->tail, s, &suffix_idx,
+                             alpha_map_char_to_trie (trie->alpha_map, *p)))
+        {
+            return FALSE;
+        }
+        if (0 == *p)
+            break;
+    }
 
     /* found, set the val and return */
     if (o_data)
@@ -148,31 +224,55 @@ trie_retrieve (Trie *trie, const TrieChar *key, TrieData *o_data)
 }
 
 Bool
-trie_store (Trie *trie, const TrieChar *key, TrieData data)
+trie_store (Trie *trie, const AlphaChar *key, TrieData data)
 {
     TrieIndex        s, t;
     short            suffix_idx;
-    const TrieChar  *p;
-    size_t           len;
+    const AlphaChar *p, *sep;
 
     /* walk through branches */
     s = da_get_root (trie->da);
     for (p = key; !trie_da_is_separate (trie->da, s); p++) {
-        if (!da_walk (trie->da, &s, *p))
-            return trie_branch_in_branch (trie, s, p, data);
-        if ('\0' == *p)
+        if (!da_walk (trie->da, &s,
+                      alpha_map_char_to_trie (trie->alpha_map, *p)))
+        {
+            TrieChar *key_str;
+            Bool      res;
+
+            key_str = alpha_map_char_to_trie_str (trie->alpha_map, p);
+            res = trie_branch_in_branch (trie, s, key_str, data);
+            free (key_str);
+
+            return res;
+        }
+        if (0 == *p)
             break;
     }
 
     /* walk through tail */
+    sep = p;
     t = trie_da_get_tail_index (trie->da, s);
     suffix_idx = 0;
-    len = strlen ((const char *) p) + 1;    /* including null-terminator */
-    if (tail_walk_str (trie->tail, t, &suffix_idx, p, len) != len)
-        return trie_branch_in_tail (trie, s, p, data);
+    for ( ; ; p++) {
+        if (!tail_walk_char (trie->tail, t, &suffix_idx,
+                             alpha_map_char_to_trie (trie->alpha_map, *p)))
+        {
+            TrieChar *tail_str;
+            Bool      res;
+
+            tail_str = alpha_map_char_to_trie_str (trie->alpha_map, sep);
+            res = trie_branch_in_tail (trie, s, tail_str, data);
+            free (tail_str);
+
+            return res;
+        }
+        if (0 == *p)
+            break;
+    }
 
     /* duplicated key, overwrite val */
     tail_set_data (trie->tail, t, data);
+    trie->is_dirty = TRUE;
     return TRUE;
 }
 
@@ -195,6 +295,7 @@ trie_branch_in_branch (Trie           *trie,
     tail_set_data (trie->tail, new_tail, data);
     trie_da_set_tail_index (trie->da, new_da, new_tail);
 
+    trie->is_dirty = TRUE;
     return TRUE;
 }
 
@@ -240,38 +341,47 @@ fail:
 }
 
 Bool
-trie_delete (Trie *trie, const TrieChar *key)
+trie_delete (Trie *trie, const AlphaChar *key)
 {
     TrieIndex        s, t;
     short            suffix_idx;
-    const TrieChar  *p;
-    size_t           len;
+    const AlphaChar *p;
 
     /* walk through branches */
     s = da_get_root (trie->da);
     for (p = key; !trie_da_is_separate (trie->da, s); p++) {
-        if (!da_walk (trie->da, &s, *p))
+        if (!da_walk (trie->da, &s,
+                      alpha_map_char_to_trie (trie->alpha_map, *p)))
+        {
             return FALSE;
-        if ('\0' == *p)
+        }
+        if (0 == *p)
             break;
     }
 
     /* walk through tail */
     t = trie_da_get_tail_index (trie->da, s);
     suffix_idx = 0;
-    len = strlen ((const char *) p) + 1;    /* including null-terminator */
-    if (tail_walk_str (trie->tail, t, &suffix_idx, p, len) != len)
-        return FALSE;
+    for ( ; ; p++) {
+        if (!tail_walk_char (trie->tail, t, &suffix_idx,
+                             alpha_map_char_to_trie (trie->alpha_map, *p)))
+        {
+            return FALSE;
+        }
+        if (0 == *p)
+            break;
+    }
 
     tail_delete (trie->tail, t);
     da_set_base (trie->da, s, TRIE_INDEX_ERROR);
     da_prune (trie->da, s);
 
+    trie->is_dirty = TRUE;
     return TRUE;
 }
 
 typedef struct {
-    Trie           *trie;
+    const Trie     *trie;
     TrieEnumFunc    enum_func;
     void           *user_data;
 } _TrieEnumData;
@@ -282,7 +392,7 @@ trie_da_enum_func (const TrieChar *key, TrieIndex sep_node, void *user_data)
     _TrieEnumData  *enum_data;
     TrieIndex       t;
     const TrieChar *suffix;
-    TrieChar       *full_key;
+    AlphaChar      *full_key, *p;
     Bool            ret;
 
     enum_data = (_TrieEnumData *) user_data;
@@ -290,9 +400,16 @@ trie_da_enum_func (const TrieChar *key, TrieIndex sep_node, void *user_data)
     t = trie_da_get_tail_index (enum_data->trie->da, sep_node);
     suffix = tail_get_suffix (enum_data->trie->tail, t);
 
-    full_key = (TrieChar *) malloc (strlen ((const char *)key)
-                                    + strlen ((const char *)suffix) + 1);
-    strcat (strcpy ((char *)full_key, (const char *)key), (const char *)suffix);
+    full_key = (AlphaChar *) malloc ((strlen ((const char *)key)
+                                      + strlen ((const char *)suffix) + 1)
+                                     * sizeof (AlphaChar));
+    for (p = full_key; *key; p++, key++) {
+        *p = alpha_map_trie_to_char (enum_data->trie->alpha_map, *key);
+    }
+    for ( ; *suffix; p++, suffix++) {
+        *p = alpha_map_trie_to_char (enum_data->trie->alpha_map, *suffix);
+    }
+    *p = 0;
 
     ret = (*enum_data->enum_func) (full_key,
                                    tail_get_data (enum_data->trie->tail, t),
@@ -303,7 +420,7 @@ trie_da_enum_func (const TrieChar *key, TrieIndex sep_node, void *user_data)
 }
 
 Bool
-trie_enumerate (Trie *trie, TrieEnumFunc enum_func, void *user_data)
+trie_enumerate (const Trie *trie, TrieEnumFunc enum_func, void *user_data)
 {
     _TrieEnumData   enum_data;
 
@@ -320,7 +437,7 @@ trie_enumerate (Trie *trie, TrieEnumFunc enum_func, void *user_data)
  *-------------------------------*/
 
 TrieState *
-trie_root (Trie *trie)
+trie_root (const Trie *trie)
 {
     return trie_state_new (trie, da_get_root (trie->da), 0, FALSE);
 }
@@ -330,7 +447,10 @@ trie_root (Trie *trie)
  *----------------*/
 
 static TrieState *
-trie_state_new (Trie *trie, TrieIndex index, short suffix_idx, short is_suffix)
+trie_state_new (const Trie *trie,
+                TrieIndex   index,
+                short       suffix_idx,
+                short       is_suffix)
 {
     TrieState *s;
 
@@ -346,6 +466,13 @@ trie_state_new (Trie *trie, TrieIndex index, short suffix_idx, short is_suffix)
     return s;
 }
 
+void
+trie_state_copy (TrieState *dst, const TrieState *src)
+{
+    /* May be deep copy if necessary, not the case for now */
+    *dst = *src;
+}
+
 TrieState *
 trie_state_clone (const TrieState *s)
 {
@@ -366,12 +493,14 @@ trie_state_rewind (TrieState *s)
 }
 
 Bool
-trie_state_walk (TrieState *s, TrieChar c)
+trie_state_walk (TrieState *s, AlphaChar c)
 {
+    TrieChar tc = alpha_map_char_to_trie (s->trie->alpha_map, c);
+
     if (!s->is_suffix) {
         Bool ret;
 
-        ret = da_walk (s->trie->da, &s->index, c);
+        ret = da_walk (s->trie->da, &s->index, tc);
 
         if (ret && trie_da_is_separate (s->trie->da, s->index)) {
             s->index = trie_da_get_tail_index (s->trie->da, s->index);
@@ -381,24 +510,26 @@ trie_state_walk (TrieState *s, TrieChar c)
 
         return ret;
     } else {
-        return tail_walk_char (s->trie->tail, s->index, &s->suffix_idx, c);
+        return tail_walk_char (s->trie->tail, s->index, &s->suffix_idx, tc);
     }
 }
 
 Bool
-trie_state_is_walkable (const TrieState *s, TrieChar c)
+trie_state_is_walkable (const TrieState *s, AlphaChar c)
 {
+    TrieChar tc = alpha_map_char_to_trie (s->trie->alpha_map, c);
+
     if (!s->is_suffix)
-        return da_is_walkable (s->trie->da, s->index, c);
+        return da_is_walkable (s->trie->da, s->index, tc);
     else 
         return tail_is_walkable_char (s->trie->tail, s->index, s->suffix_idx,
-                                      c);
+                                      tc);
 }
 
 Bool
-trie_state_is_leaf (const TrieState *s)
+trie_state_is_single (const TrieState *s)
 {
-    return s->is_suffix && trie_state_is_terminal (s);
+    return s->is_suffix;
 }
 
 TrieData
