@@ -16,13 +16,14 @@
  */
 
 /**
- * @brief Trie IO modes
+ * @brief Trie character type for alphabet
  */
-typedef enum {
-    TRIE_IO_READ   = 0x01,
-    TRIE_IO_WRITE  = 0x02,
-    TRIE_IO_CREATE = 0x04
-} TrieIOMode;
+typedef uint32         AlphaChar;
+
+/**
+ * @brief Error value for alphabet character
+ */
+#define ALPHA_CHAR_ERROR   (~(AlphaChar)0)
 
 /**
  * @brief Trie character type for key
@@ -37,7 +38,7 @@ typedef unsigned char  TrieChar;
 /**
  * @brief Type of Trie index
  */
-typedef int16          TrieIndex;
+typedef int32          TrieIndex;
 /**
  * @brief Trie error index
  */
@@ -45,12 +46,12 @@ typedef int16          TrieIndex;
 /**
  * @brief Maximum trie index value
  */
-#define TRIE_INDEX_MAX    32767
+#define TRIE_INDEX_MAX    0x7fffffff
 
 /**
  * @brief Type of value associated to trie entries
  */
-typedef int16          TrieData;
+typedef int32          TrieData;
 /**
  * @brief Trie error data
  */
